 #!/usr/bin/perl

# ===================================================================================================
# create dir names using the ncbi_name and store the plasmid DNA data there using the s.identifier
# NC_xxx_name_PLASMID<0#>.fna
#
# Note: each file will contain ONLY the direct-strand DNA; scripts that use this data
# for instance to do word counting may need to extend with the reverse complement
# ===================================================================================================

use strict;
use warnings;

use DBI;
use File::Spec;
use Data::Dumper;

sub overlap($\@);

# ---------- main() ------------

my $STAT_FILE = "Chr_STATS.xls";
open( my $STAT, '>', $STAT_FILE ) or do { die $!; };
print $STAT "CHR#\tNCBI NAME\tLength (bp)\tProportion Coding\tNumber of Genes\tTotal number of Overlapping Pairs of Genes\t";
print $STAT "Overlap_Complete\tOverlap_Partial\tOverlap_OR\n";

# global arrays to handle ptt info for each genome
# we'll sort indices of these arrays (in index) to reflect sorted starting bp
my @indices;
my @start;
my @stop;
my @accession;
my @strand;

run();

# ------------------------------


sub run
{
	#create folder to store data in unless folder already exists
	my $rootDir = "data_all_chr";
        mkdir $rootDir unless (-d $rootDir);

	# fetch handle to dB
	my $dbh	= mysql_dbh();

	# SELECT all ORGANISMS
	my $sth1	= $dbh->prepare( <<"END" );

	SELECT
		m.ncbi_dir, m.project_id
	FROM
		metadata m;
END

	print "\n\nSELECT all organisms ...\n";
	$sth1->execute( );
	print "\nall organisms query DONE.\n\n";
	my $ref_seq;

	my $n=1; #

	# for each ORGANISM
	while ( my $data = $sth1->fetchrow_hashref ) #fetch reference to a hash table
        					     #containing a row of data
	{

#xxxxxxxxx HACK to test
#if ($n > 100){
#	$n++;
#	next;
#}
#print "n is: $n \n";
#xxxxxxxxx HACK to test
#if ($n != 96) {
#	$n++;
#	next;
#}
#xxxxxxxxx HACK to test

		my $ncbi_name = $data->{'ncbi_dir'};
		my $orgID     = $data->{'project_id'};  #orgID is an ID number identifying
                					#the organism

#xxxxxxxxx HACK to test
#if ($orgID != 12986) {
#	$n++;
#	next;
#}

		# parse off leading stuff:  ftp://ftp.ncbi.nih.gov/genomes/Bacteria/
                # this will get the name of the organism
		$ncbi_name =~ m#.*\/(.+?)$#;
		$ncbi_name = $1;

		#print "bug name:  $only_bug_name\n";
		#next;

		my $basedir	= File::Spec->catdir( 'data_all_chr', $ncbi_name );
		#mkdir $basedir unless (-d $basedir);

		# get the chromosomes from each organism
		my $sth2	= $dbh->prepare( <<"END2" );

		SELECT
			s.number, s.ref_seq, s.sequence
		FROM
			sequences s
		WHERE
			s.project_id = ?
			AND s.type="Chromosome";
END2

		#print "\n\nSELECT all chrs ...\n";
		$sth2->execute( $orgID ); #get the dna sequence of current organism
		#print "\nall chrs query DONE.\n\n";

		# for each CHR
		while ( my $chr_data = $sth2->fetchrow_hashref )#fetch a refrence to a hash
                                                                #containing information about
                                                                #chromosome
		{
			#my $sequenceID      = $chr_data->{'id'};
			$ref_seq       		 = $chr_data->{'ref_seq'}; #?
			my $sequenceNUMBER 	 = $chr_data->{'number'};  #?

			my $dna   = $chr_data->{'sequence'};

			my $lengthDNA = length($dna);

			if ($lengthDNA <= 0) {
				print "$ncbi_name has no DNA sequence in dB; organism ignored ...\n";
				next;
			}

			#---------------------------------------------------------------------------------
			# grab the protein table (for stats purposes only at this point)
			my $sth3	= $dbh->prepare( <<"END3" );

			SELECT
			 	p.accession, p.start, p.stop, p.strand
			FROM
			 	proteins p
			WHERE
				p.ref_seq = ?;
END3

			$sth3->execute( $ref_seq ); #get protein table for specific chromosome
			#print "\nptt dB query DONE.\n\n";

			my $totalGenes = 0;				# sum total of the number of genes in this plasmid

			my $previousEnd   =  0;			# to enable detection/count of the number of overlapping genes
			my $previousStart = -1;
			my $totalOverlappingGenes = 0;

                        #keep counts for different types of overlaps. Look at the comments
                        #inside the upcoming if block for definitions of each type
			my $overlap_partial = 0;
			my $overlap_complete = 0;
			my $numOverlapOR=0;

			my $totalCoding = 0;			# sum total of coding regions

                        my @allGenes;				#array holding all genes

			while ( my $ptt = $sth3->fetchrow_hashref )#get a hash refrence containing
                        					   #information about the protein/gene
			{
				$totalGenes++;

				my $start     = $ptt->{'start'};
				my $stop      = $ptt->{'stop'};
				my $accession = $ptt->{'accession'};
				my $strand    = $ptt->{'strand'}; #which strand of dna
                                				  #is the gene on

                                #store the gene's start and stop information
                                my @Current = ($start, $stop);   #create an array holding the start and stop positions

                                my $CurrentRef = \@Current;      #since we need to store a two element array in an array create a reference
                                push(@allGenes, $CurrentRef);    #add gene to array of genes




			} # while more gene entries in .ptt file

                        ($overlap_partial, $overlap_complete, $numOverlapOR, $totalCoding) = overlap($lengthDNA, @allGenes); #run the overlap subroutine
                        $totalOverlappingGenes = $overlap_partial + $overlap_complete; #count the total number of overlaps by taking the sum of the two types


			my $percentCoding;

			my $ptt_size = $totalGenes;
			if ($ptt_size <= 0)
			{
				$percentCoding = 0;
				printf "WARNING (no ptt table in dB): CHR_%d %s\n", $sequenceNUMBER, $ncbi_name;
				next;
			}

			$percentCoding = $totalCoding/$lengthDNA;
			#---------------------------------------------------------------------------------




			mkdir $basedir unless (-d $basedir);

			# establish filename for the respective directory
			#my $NCfilename = $chr_data->{'ref_seq'};
			my $file	= File::Spec->catfile( $basedir, sprintf( "%s_chr%02d.fna", $ref_seq, $sequenceNUMBER ) );

			if (-f $file) {
				warn "File already exists: $file\n";
				next;
			}
			else
			{
				# add statistics on this bug to the stats file
				printf $STAT "CHR_%d\t%s\t%d\t%10.3f\t%d\t%d\t", $sequenceNUMBER, $ncbi_name, $lengthDNA,
														   $percentCoding, $totalGenes, $totalOverlappingGenes;
				printf $STAT "%d\t%d\t%d\n", $overlap_complete, $overlap_partial, $numOverlapOR;

				# print DNA
				printf "[%03d] CHR_%d %s\n", $n, $sequenceNUMBER, $ncbi_name;

				open( my $fh, '>', $file ) or do { warn $!; next };

				#print { $fh } $chr_data->{'ncbi_directory'};
				#print { $fh } $chr_data->{'dna'};
				print { $fh }  $dna;

				close($fh);
			}
		 } # while more CHRs in this organism

		 $n++;

	} # while more ORGANISMS

	close ($STAT);
} # end sub run




#------------------------------------------------------------------
# return connection handle to dB

# Replace the information here in order to access your database
# Use Seed.pl to fill your own database with NCBI genomes
sub mysql_dbh {
	my $db		= 'test';
        my $host        = 'localhost';

	my $user	= 'GenomicsUser';
	my $pass	= '';

        my $dbh = DBI->connect ( "DBI:mysql:database=$db;host=$host", $user, $pass, { RaiseError => 1 } );

	return $dbh;
}

#------\
#overlap\
#-------------------------------------------------------------------------------
# Summary: Takes a list of gene positions and counts the number of overlaps between said
# genes, as well as the number of base pairs within the genes avoiding any double counting
#
# In:	   2 arguments: -length of DNA sequence
#                       -reference to an array of refrences to two item arrays
#                       each of which contains the start location of a gene in slot 0 and the
#			stop location in slot 1. This array should contain all the genes in the sequence
#
# Out:	   4 values:	-the number of pairs of genes which partially overlap
#			-the number of pairs of genes in which one gene is completly contained within another
#                       -the number of genes that pass through the OR
#			-the total number of unique (no double counting) base pairs within all the gene
#
# Algorithm: To detect overlaps this algorithm takes all the genes and sorts them by their starting position. It also keeps
# track of a list of genes to compare a gene with to check for overlaps. The subroutine goes through all the genes in order
# of their starting location comparing them against all the genes in the list looking for overlaps and adds them to the
# list when the comparisions are done. When a gene in the list ends before the current gene starts it is removed from the
# list as do to the genes being sorted in order of starting location all the genes after the current gene will also
# start after the end of that gene, meaning the no more overlaps are possibly. By removing genes from the list to compare
# like this the number of comparisions is minimized

sub overlap ($\@){

	my $START = 0;                  #location of the start position in the array holding a gene
	my $STOP = 1;                   #location of the stop position in the array holding a gene
	my $partialoverlaps = 0;        #number of pairs of genes which partially overlap eachother
	my $completeoverlaps = 0;       #number of pairs of genes in which one gene is completely inside another
	my $totalORoverlaps = 0;        #genes that overlap the OR
	my $length;                     #length of gene
	my $allGenesLocRef;

	#these two values give the portion of the chromosome which has already had the base pairs
	#in genes counted so as to avoid double counting
	my $startRegionCounted = -1;     #where is the start of the region that has been counted
	my $endRegionCounted = -1;       #where is the end of the region that has been counted
	my $endOverlapRegionCounted = 0; #how far has the count gone past the OR when dealing with genes
	                                  #the pass the OR

	my $totalBPInGenes = 0;          #how many base pairs total are there in all the genes without any doublecounting
	                                 #from overlapping genes

	my @allGenesLoc;                         #array holding all the genes in chromosome
	my @nonPassed;                   #array holding all genes that still need to have comparisions run against
	                                 #for overlaps. This is all genes that end after the start of the current gene

	($length, $allGenesLocRef) = @_; #get the pramaters
	@allGenesLoc =@{$allGenesLocRef};

	@allGenesLoc = sort{@{$a}[$START] <=>  @{$b}[$START]} @allGenesLoc;  #sort the genes by starting position from lowest to highest

	$startRegionCounted = $allGenesLoc[0][$START]; #mark the region that has been counted as starting with the first gene
	                                            #this is necessary to avoid double counting with genes that overlap the OR

	foreach my $curRef(@allGenesLoc)                    #go through every gene
	{
	        my @toDelete; #holds index numbers of elements to remove from array
	        my $deleteOffset = 0; #since each deletion will change index numbers for rest of array
	                              #need an offset to find right elements.

	        my @curGene = @{$curRef}; #current gene being checked for overlaps. Gotten by derefrencing the refrence in @allGenesLoc
	        my $ORoverlap = 0;       #does this gene overlap the OR

	        if ($curGene[$STOP] < $curGene[$START]) #check if current gene overlaps the OR
	        {
	                $totalORoverlaps++;
	                $ORoverlap = 1;                 #mark that gene overlaps OR
	        }

	        if((scalar @nonPassed) > 0) #if there are genes in nonPassed
	        {
	                for my $i(0 ..(scalar @nonPassed - 1))  #go through every gene in nonPassed
	                                                        #these genes will be refered to as the second gene
	                {
	                    my @secondGene = @{$nonPassed[$i]}; #get and derefrence gene to compare to

	                    if ($secondGene[$START] < $secondGene[$STOP]) #if second
	                                                                #does not overlap OR
	                    {
	                            if ($curGene[$START] >  $secondGene[$STOP]) #if current gene starts
	                                                                      #after second gene ends
	                            {
	                                push (@toDelete, $i); #second gene is no longer needed so add to list to delete
	                            }

	                            else                                #otherwise there is an overlap
	                                                                #as the current gene starts before the second Gene
	                                                                #ends
	                            {
	                                if(!$ORoverlap)                 #if current gene does not
	                                                                #overlap OR
	                                {
	                                        if($curGene[$STOP] <= $secondGene[$STOP]) #if the current gene also ends
	                                                                                  #before the current gene starts
	                                        {
	                                                $completeoverlaps++;              #the current gene is completly within
	                                                                                  #the second gene
	                                        }
	                                        else                                      #otherwise
	                                        {
	                                                $partialoverlaps++;               #the current gene only partially overlaps
	                                                                                  #the second gene
	                                        }
	                                }

	                                else
	                                {
	                                        $partialoverlaps++; #it is impossible for an gene overlaping OR
	                                                            #to completley overlap a gene that doesn't
	                                }
	                            }
	                    }

	                    else                #second gene overlaps OR
	                                        #there must be an overlap, because the second gene starts before the current gene
	                                        #due to the array of all genes being sorted by starting point from earliest to latest
	                                        #and the second gene must come from before the current gene in this array
	                                        #and the second gene ends after the current genes start since overlapping the OR means that
	                                        #it contains every base pair between it's start and the OR which would be every position with
	                                        #a higher number then it's start, which includes as mentioned above the current gene

	                    {
	                        if(!$ORoverlap) #if current gene doesn't overlap OR
	                        {
	                                $completeoverlaps++; #since current gene starts after the second gene(
	                                                     #and doesn't pass the OR, it has to be entirely within
	                                                     #the second gene which does passs over the OR
	                        }

	                        else            #current gene overlaps OR
	                        {
	                                if($curGene[$STOP] <= $secondGene[$STOP])#if the current gene stops before the second gene
	                                {
	                                        $completeoverlaps++; #the current gene is contained completly within the second gene
	                                                             #as is starts after and ends before the second gene
	                                }
	                                else                                     #otherwise the current gene stops after the second gene
	                                {
	                                        $partialoverlaps++;              #therefore there is only a partial overlap
	                                }
	                        }
	                    }
	                }
	        }

	        if($ORoverlap)                  #if the current gene overlaps the OR
	        {

	                #we need to check if the gene overlaps any genes at the begining of the genome which should
	                #have already been removed from the list of genes to compare to.

	                my $pastStop = 0;       #has the first gene after the end of the current gene been found
	                my $k = 0;              #index number of gene currently being compared with

	                while(!$pastStop && $k < (scalar @allGenesLoc) - 1)     #go through array until the first gene that starts past the end
	                                                                        #of the current gene is found or the end of the array is reached
	                {
	                        if($allGenesLoc[$k][$START] <= $curGene[$STOP]) #if the gene being read from the array starts before the current gene stops
	                                                                        #there is overlap
	                        {
	                                if($allGenesLoc[$k][$STOP] <= $curGene[$STOP]) #if the gene read from the array also stops before the current gene does
	                                {
	                                        $completeoverlaps++;                #this gene is completely within the current gene
	                                }
	                                else                                        #otherwise the gene read from the array stops after the current gene does
	                                {
	                                        $partialoverlaps++;                 #there is only a partial overlap
	                                }
	                        }

	                        else         #otherwise the gene read from the array starts after the current gene stops
	                        {
	                                $pastStop = 1;           #this means we've gone past any genes that might overlap so mark that
	                        }

	                        $k++;        #go to next gene
	                }
	        }

	        #handle counting the number of base pairs within the genes
	        #the basic idea is that two variables $startRegionCounted, and $endRegionCounted mark the two ends
	        #of the portion of the genome which has already had the base pairs in a gene counted. Only the basepairs
	        #in the current gene that do not lie within this range will get added to the total

	        if(!$ORoverlap) #if the current gene doesn't overlap
	        {
	                if ($curGene[$START] > $endRegionCounted) #if the start comes after the end
	                                                          #of the region encountered we just
	                                                          #skipped an intergenic region
	                {
	                    $totalBPInGenes += ($curGene[$STOP] - $curGene[$START] + 1); #add number of BP in entire gene
	                    $endRegionCounted = $curGene[$STOP];        #move the end of the region counted
	                                                                #to the end of the current gene
	                }
	                elsif ($curGene[$STOP] > $endRegionCounted)#if there is a partial overlap going over the end
	                                                          #of a previous gene
	                {
	                    $totalBPInGenes += ($curGene[$STOP] - $endRegionCounted);    #add all bp after end of region already counted
	                    $endRegionCounted = $curGene[$STOP];        #move the end of the region counted
	                                                                #to the end of the current gene
	                }
	        }
	        else #current gene Overlaps the OR
	        {
	             if ($curGene[$STOP] > $allGenesLoc[0][$STOP]) #if the current gene goes past the end of the first gene in the array
	                                                        #the code can not currently ensure a correct count
	             {
	                print STDERR "Current gene overlaps the OR and continues past the end of the first gene. Unable to ensure accurate count. Exiting now\n";
	                exit;
	             }

	             if ($endRegionCounted != $length)          #if the end of the region counted is not equal to the length of
	                                                        #the genome
	             {
	                #since any gene that overlaps the OR will be definition have to cover the region between it's start and the
	                #furthest part of the genome add that entire portion to the count

	                if($curGene[$START] > $endRegionCounted) #if the curGene starts after the end of the region counted
	                                                         #there is an intergenic region between the end of the region
	                                                         #counted and the current gene
	                {
	                        $totalBPInGenes += ($length - $curGene[$START] + 1); #get number of base pairs between the start of
	                                                                             #the gene and the OR
	                }

	                else                                    #otherwise curGene partially overlaps the previous gene
	                {
	                       $totalBPInGenes += ($length - $endRegionCounted);     #get the number of genes from the end
	                                                                             #of the region coutned to the end of the
	                                                                             #genome
	                }

	                $endRegionCounted = $length; #mark that we've counted all the base pairs up to the end of the genome
	             }  #end if $endRegionCounted != $length

	             if ($curGene[$STOP] > $endOverlapRegionCounted && $endOverlapRegionCounted < $startRegionCounted - 1)
	             #if the current gene stops after the farthest end of any previous gene that overlaps the OR
	             #and if the region counted for base pairs in genes overlapping the OR hasn't reached the first gene
	             #basically is there any base pairs left to count at the start of the genome
	             {
	                if($curGene[$STOP] < $startRegionCounted) #if the stop point is before the startRegion counted
	                                                          #theres a gap between this gene and the first gene in the
	                                                          #genome
	                {
	                        $totalBPInGenes += ($curGene[$STOP] - $endOverlapRegionCounted); #number of base pairs is the number
	                                                                                         #of pairs between the end of the region
	                                                                                         #counted and the end of the gene
	                        $endOverlapRegionCounted = $curGene[$STOP];
	                }
	                else                                      #otherwise the current gene overlaps the first gene in the genome
	                {
	                        $totalBPInGenes += ($startRegionCounted - $endOverlapRegionCounted - 1); #get number of base pairs between
	                                                                                                 #end of region counted and start of first gene
	                        $endOverlapRegionCounted = $startRegionCounted;
	                }
	             }  #end if there are base pairs to count after OR
	        } #end if current gene overlaps OR

	        foreach my $j(@toDelete) #go through each element to delete
	        {
	                splice(@nonPassed, ($j - $deleteOffset), 1);
	                $deleteOffset++; #keeping track of an offset here to account for the fact that the array is shrinking in size
	                                 #since the index numbers to delete were generated by a for loop going through the array
	                                 #we can be sure that they are stored in @toDelete in order from lowest to highest so each deletion
	                                 #will always lower the index number of the remaining elements to delete by 1
	        }

	        push(@nonPassed, $curRef); #add current gene to list of genes to compare against
	}

        return($partialoverlaps, $completeoverlaps, $totalORoverlaps, $totalBPInGenes);

}

__END__

head1 = NAME

C<extract_All_chrs.pl>

=head1 SYNOPSIS

% perl extract_ALL_chrs.pl

The packages DBI and DBD_mySql must be installed

This script must currently be run through an account logged onto the lexomics server

This will likely change soon

=head1 DESCRIPTION

=head2 Summary

This script accesses a database and retrieves all the different organisms on the server
and gets some basic information about them

=head2 Input

The script needs access to a mysql server containing the genomes of a number of organisms
and their metadata. If you have already have a mysql server the script seed.pl can be used
to create and populate the database with all the micro-organisms with complete genomes
in NCBI's database.

=head3 Connecting to Your Database

Once you have a database ready you will need to make a few minor edits to the script so
it can connect to your database. Search this file for the mysql_dbh subroutine:

 sub mysql_dbh {

Modify these four lines, replacing this generic data with your database's access info:

 my $db         = 'test';
 my $host       = 'localhost';
 my $user       = 'GenomicsUser';
 my $pass       = '';

For $db enter the name of the (MySQL) database you are using.
   Ex: 'GenomeDatabase'

For $host enter the name or address of the server your database is on.
   Ex: 'WheatonGenomics'

For $user enter your MySQL username on the database.
   Ex: 'wsmith'

For $pass enter the user's corresponding password (or leave it blank).
   Ex: 'lollipop'

=head2 Output

Every chromosome for every organism will be stored in a seperate .fna file in a subdirectory
in the data_all_chr folder named after the organism. In addition in the folder the script is run in
a file Chr_Stats.xls will contain for every chromosome in every organism the length, percentage coding
the number of genes, the number of overlaps and the number of each type of overlap

=head3 Overlap codes

 Complete - one gene completely within a second gene
 	|----------------|
     	      |----|

 Partial one gene partially inside second gene
         |-------|
              |----|

 (OR) the gene overlaps the origin of replication

=head1 AUTHORS

 Mark D. LeBlanc
 Donald W. Bass

=head1 Modification History

=head3 6/17/2010 (nkf)

Removed Wheaton database log-in info from the mysql_dbh subroutine.
Users must replace the generic data with their own database access info.

=head3 6/15/2010 (dwb)

Made script create the folder data_all_chr if it does not already exist

=head3 6/07/2010 (dwb)

Completely rewrote code for detecting overlaps to make it more accurate
and also simplified the number of types of overlaps from 5 to 2, by removing two
types that were mirror images of another two types, and removing overlap type 5
which stood for no overlap, and didn't make sense with the new code which instead
of running one comparision per gene, runs as many as necessary to catch all overlaps

=head3 6/03/2010  (dwb)

added additional code documentation

=head3 6/02/2010 (dwb)

added pod documentation

=head3 12/02/2008 (mdl)

worked on genic-intergenic STAT output to get a better feel for the
extent of genic regions, including the types of operonds and/or
overlapping regions between consecutive genes (see ChrSTAT.xls output)

=head3 11/20/2008 (mdl)

doing only CHROMOSOMES now

=head3 12/03/2007 (mdl)

REMOVED abstracted_chromosome code ...
created a file for stats of plasmids

=head3 11/30/2007 (mdl)

looked again at why length of abs-chr is not equal to length of
real chr + extras;  NOT SOLVED ...

=head3 10/19/2007 (mdl)

start morphing to work with PLASMIDS

=head3 02/01/2007 (mdl)

dang!  some genomes have a final gene that bridges the (man-made) origin (bp 1);
inserting fix (and code to check if longer abstracted chromosome length is right

=head3 # 01/29/2007 (mdl) -

ABSTRACTED CHROMOSOME
genes on the indirect(-) strand are stored in their reverse complement form
so that all counting will be in a 5' to 3' direct
all gene regions are stored independently even if two
gene regions overlap;

 5'  ...|---1---|......      3'

 3' ...|---2---|....         5'

these two regions will be completely consecutively  #1.N.#2'
where 2' means the sequence in #2 is stored in reverse complement fashion;

=head3 08/22/2006 (mdl) -

modified to handle multiple chromosomes per genome

=head4 NOTE: 10/19/2007

(should also handle multiple plasmids)

=head3 07/13/2006 (mdl)

mostly done

=head1 COPYRIGHT INFORMATION

 ==========================================================================
   Copyright (C) 2010  Wheaton Genomics Research Group, Norton MA

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

#what is 562